import logging.config

# 日志类型（若配置为consoleLogger则输出到控制台，若配置为fileLogger则记录到日志文件中）
LOGGER_TYPE = 'fileLogger'
# LOGGER_TYPE = 'consoleLogger'

# 配置全局Logger
configFile = open('Core/Config/SpiderLoggingConfig.conf', encoding='utf8')
logging.config.fileConfig(configFile)
log = logging.getLogger(LOGGER_TYPE)



import requests
import re
import logging
import time
from Core.Logger import log

# 代理 IP 验证网站
url = 'http://icanhazip.com/'
# 请求头信息
header = {'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,*/*;q=0.8',
          'Accept-Encoding': 'gzip, deflate, sdch',
          'Accept-Language': 'zh-CN,zh;q=0.8',
          'Cache-Control': 'max-age=0',
          'Host': 'www.icanhazip.com',
          'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) '
                        'Chrome/56.0.2924.87 Safari/537.36'}

# 连接超时设置（单位：秒）
CONNECT_TIMEOUT = 30
# 连接重试次数
NETWORK_RECONNECT_TIMES = 3


class DataValidateModule:
    def __init__(self):
        self.session = requests.session()

    # 验证给定的代理 IP 是否有效
    def validate_proxy_ip(self, proxy_ip_info):
        from Proxy import proxyCore
        if proxy_ip_info is None:
            return False

        # 构造代理信息
        proxy_ip = proxy_ip_info[proxyCore.PROXY_IP]
        proxy_port = proxy_ip_info[proxyCore.PROXY_PORT]
        proxy_protocol = proxy_ip_info[proxyCore.PROXY_PROTOCOL].lower()
        proxy = {proxy_protocol: proxy_ip + ':' + proxy_port}

        # 使用代理进行连接
        retry_time = 0
        while retry_time < NETWORK_RECONNECT_TIMES:
            try:
                response = requests.get(url, timeout=CONNECT_TIMEOUT, headers=header, proxies=proxy)

                # 解析返回的当前使用的IP并判断是否有效
                match_list = re.findall(r'[0-9]+(?:\.[0-9]+){3}', response.text)
                if len(match_list) > 0:
                    current_ip = match_list.pop()
                    if current_ip is not None and current_ip == proxy_ip:
                        if log.isEnabledFor(logging.DEBUG):
                            log.debug("获取到一个可用的代理IP")
                        return True
                else:
                    retry_time += 1
                    time.sleep(1)
            except Exception:
                retry_time += 1
        return False



from Proxy import validateData
from Proxy import fetchData
from Proxy import parseData
from Core.Logger import log
import time
import threading
import queue
import configparser
import logging

# 代理信息键值字段
PROXY_ID = 'id'
PROXY_IP = 'ip'
PROXY_PORT = 'port'
PROXY_ADDRESS = 'address'
PROXY_PROTOCOL = 'protocol'
PROXY_ALIVE_TIME = 'aliveTime'
PROXY_VALIDATE_TIME = 'validateTime'

# 代理网页检索起始页
FETCH_START_PAGE = 1
# 代理网页检索最大截至页
FETCH_END_PAGE = 5

# 代理池大小
PROXY_POOL_SIZE = 10
# 代理池扫描更新间隔
PROXY_POOL_SCAN_INTERVAL = 300
# 代理验证线程数
PROXY_VALIDATE_THREAD_NUM = 3

# 待验证的代理信息列表
unchecked_proxy_list = queue.LifoQueue(300)
# 可用的代理池
proxy_pool = queue.Queue(100)

# 标志量，是否正在扫描代理池
is_scanning = False


# 代理服务守护线程
class ProxyDaemonThread(threading.Thread):
    def __init__(self):
        threading.Thread.__init__(self)

    def run(self):
        # 初始化配置
        self.init()

        # 启动代理检验线程
        validate_thread_list = []
        for i in range(PROXY_VALIDATE_THREAD_NUM):
            validate_thread = ProxyValidateThread()
            validate_thread_list.append(validate_thread)
            validate_thread.start()
        if log.isEnabledFor(logging.DEBUG):
            log.debug("代理验证线程启动")

        # 启动代理池扫描线程
        scan_thread = ProxyPoolScanThread()
        scan_thread.start()
        if log.isEnabledFor(logging.DEBUG):
            log.debug("代理池扫描线程启动")

        # 检查是否有线程出现异常并将其重启
        while True:
            # 检查代理验证线程
            for thread in validate_thread_list:
                if thread.status == 'error':
                    validate_thread_list.remove(thread)
                    thread = ProxyValidateThread()
                    validate_thread_list.append(thread)
                    thread.start()
                    if log.error(logging.ERROR):
                        log.error('代理验证线程重新启动')

            # 检查代理池扫描线程
            if scan_thread.status == 'error':
                scan_thread = ProxyPoolScanThread()
                scan_thread.start()
                if log.isEnabledFor(logging.ERROR):
                    log.error("代理池扫描线程重新启动")

            time.sleep(180)

    # 初始化，读取配置文件并配置
    @staticmethod
    def init():
        section = "proxy_core"
        config = configparser.ConfigParser()
        config.read('Proxy/Config/proxyConfiguration.conf', encoding='utf8')

        # validateData配置
        validateData.CONNECT_TIMEOUT = int(config.get(section, "proxyValidate_connectTimeout"))
        validateData.NETWORK_RECONNECT_TIMES = int(config.get(section, "proxyValidate_networkReconnectTimes"))
        # fetchData配置
        fetchData.CONNECT_TIMEOUT = int(config.get(section, "dataFetch_connectTimeout"))
        fetchData.NETWORK_RECONNECT_INTERVAL = int(config.get(section, "dataFetch_networkReconnectInterval"))
        fetchData.NETWORK_RETRY_TIMES = int(config.get(section, "dataFetch_networkReconnectionTimes"))
        # proxyCore配置
        global FETCH_START_PAGE
        global FETCH_END_PAGE
        global PROXY_POOL_SIZE
        global PROXY_POOL_SCAN_INTERVAL
        global PROXY_VALIDATE_THREAD_NUM
        FETCH_START_PAGE = int(config.get(section, "proxyCore_fetchStartPage"))
        FETCH_END_PAGE = int(config.get(section, "proxyCore_fetchEndPage"))
        PROXY_POOL_SIZE = int(config.get(section, "proxyCore_proxyPoolSize"))
        PROXY_POOL_SCAN_INTERVAL = int(config.get(section, "proxyCore_proxyPoolScanInterval"))
        PROXY_VALIDATE_THREAD_NUM = int(config.get(section, "proxyCore_proxyValidateThreadNum"))


# 代理检验线程
class ProxyValidateThread(threading.Thread):
    def __init__(self):
        threading.Thread.__init__(self)
        self.status = 'running'
        # 创建代理验证实例
        self.dataValidateModule = validateData.DataValidateModule()

    def run(self):
        try:
            while True:
                # 若正在扫描代理池，则暂停
                while is_scanning:
                    time.sleep(3)

                if proxy_pool.qsize() < PROXY_POOL_SIZE and unchecked_proxy_list.qsize() > 0:
                    unchecked_proxy = unchecked_proxy_list.get()
                    is_available = self.dataValidateModule.validate_proxy_ip(unchecked_proxy)
                    if is_available is True:
                        proxy_pool.put(unchecked_proxy)
                        # print(unchecked_proxy)
                    time.sleep(1)
                else:
                    time.sleep(5)
        except Exception as e:
            if log.isEnabledFor(logging.ERROR):
                log.exception(e)
            self.status = 'error'


# 代理池扫描线程，去除代理池中不可用的代理
class ProxyPoolScanThread(threading.Thread):
    def __init__(self):
        threading.Thread.__init__(self)
        self.status = 'running'
        self.current_page = 1
        # 创建数据抓取模块
        self.dataFetchModule = fetchData.DataFetchModule()
        # 创建数据解析模块
        self.dataParseModule = parseData.DataParseModule()

    def run(self):
        try:
            while True:
                if proxy_pool.qsize() < PROXY_POOL_SIZE and unchecked_proxy_list.qsize() < PROXY_POOL_SIZE:
                    self.fetch_and_parse_proxy()
                elif proxy_pool.qsize() == PROXY_POOL_SIZE:
                    if log.isEnabledFor(logging.DEBUG):
                        log.debug('代理池更新')
                    self.scan_proxy_pool()
                    time.sleep(PROXY_POOL_SCAN_INTERVAL)
                else:
                    time.sleep(60)
        except Exception as e:
            if log.isEnabledFor(logging.ERROR):
                log.exception(e)
            self.status = 'error'

    # 扫描代理池中的代理
    @staticmethod
    def scan_proxy_pool():
        # 由于待验证线程是先进后出队列，故对代理池进行扫描只需要将其添加到未检查列表，
        # 由代理检验线程对其重新验证并加入回代理池

        global is_scanning
        is_scanning = True
        while proxy_pool.qsize() > 0:
            unchecked_proxy_list.put(proxy_pool.get())
        is_scanning = False

    # 爬取并解析代理
    def fetch_and_parse_proxy(self):
        if self.current_page > FETCH_END_PAGE:
            self.current_page = FETCH_START_PAGE

        response_data = self.dataFetchModule.fetch_proxy_data(self.current_page)
        proxy_data = self.dataParseModule.parse_data(response_data)
        self.current_page += 1

        # 将解析到的代理添加到待验证的代理列表
        for proxy in proxy_data:
            unchecked_proxy_list.put(proxy)


class ProxyService:
    def __init__(self):
        self.proxy_daemon_thread = ProxyDaemonThread()

    @staticmethod
    def get_proxy():
        return proxy_pool.get()

    def start_proxy_service(self):
        self.proxy_daemon_thread.start()


# if __name__ == '__main__':
#     proxyService = ProxyService()
#     while True:
#         print(proxyService.get_proxy())



import threading
import time
import logging
from Core.Logger import log


# URL 队列调度
class Scheduler(threading.Thread):
    __slots__ = ('redis_connection', 'url_rate', 'follow_info_url_queue', 'user_info_url_queue', 'url_queue_name')

    def __init__(self, redis_connection, url_rate):
        threading.Thread.__init__(self)
        # 设置Redis连接
        self.redis_connection = redis_connection
        # 设置 url 的调度比例
        self.url_rate = url_rate

        # Following & Follower URL 队列名称
        self.follow_info_url_queue = 'followInfoURLQueue'
        # User info URL 队列名称
        self.user_info_url_queue = 'userInfoURLQueue'
        # 待下载URL队列名称
        self.url_queue_name = 'urlQueue'

        if log.isEnabledFor(logging.INFO):
            log.info('Scheduler 模块初始化完毕')

    def run(self):

        if log.isEnabledFor(logging.INFO):
            log.info('Scheduler 模块启动成功')

        while True:
            # 当 urlQueue 队列中元素太多时，停止放入
            while self.redis_connection.llen(self.url_queue_name) > 500:
                time.sleep(180)

            # 当队列中均没有元素时，暂停添加
            follow_info_queue_length = self.redis_connection.llen(self.follow_info_url_queue)
            user_info_queue_length = self.redis_connection.llen(self.user_info_url_queue)
            if follow_info_queue_length == 0 and user_info_queue_length == 0:
                time.sleep(20)
                continue

            # 分别从两个队列中获取设定比例的数量的元素添加到下载URL队列
            for i in range(self.url_rate):
                url_info = self.redis_connection.lpop(self.user_info_url_queue)
                if url_info is not None:
                    self.redis_connection.rpush(self.url_queue_name, url_info)
                    del url_info

            for i in range(10 - self.url_rate):
                url_info = self.redis_connection.lpop(self.follow_info_url_queue)
                if url_info is not None:
                    self.redis_connection.rpush(self.url_queue_name, url_info)
                    del url_info



import requests
import time

# 请求网页的 UEL
requestUrl = "http://www.xicidaili.com/nn/"

# 请求首部信息
requestHeader = {"Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,*/*;q=0.8",
                 "Accept-Encoding": "gzip, deflate, sdch",
                 "Accept-Language": "zh-CN,zh;q=0.8",
                 "Host": "www.xicidaili.com",
                 "User-Agent": "Mozilla/5.0 (Windows NT 10.0; WOW64) AppleWebKit/537.36 "
                               "(KHTML, like Gecko) Chrome/56.0.2924.87 Safari/537.36"}

# 连接重试次数
NETWORK_RETRY_TIMES = 3
# 连接重试间隔(单位：秒)
NETWORK_RECONNECT_INTERVAL = 30
# 连接超时设置（单位：秒）
CONNECT_TIMEOUT = 30


class DataFetchModule:
    def __init__(self):
        # 当前连接
        self.session = requests.session()

    # 获取代理信息
    def fetch_proxy_data(self, page):
        # 设置连接信息
        self.session.headers = requestHeader

        # 构造请求 URL
        url = requestUrl + str(page)

        # 获取数据
        retry_time = 0
        while retry_time < NETWORK_RETRY_TIMES:
            try:
                response = self.session.get(url, timeout=CONNECT_TIMEOUT)
                return response.text
            except Exception:
                # 网络异常
                # print("[代理模块]网络异常，代理信息获取失败")
                retry_time += 1
        time.sleep(NETWORK_RECONNECT_INTERVAL)

        # 超过重试次数返回 None
        return None




import configparser
import logging
import redis
import time
import pymysql
import gc
from Core.Logger import log
from Core.Downloader import Downloader
from Core.Scheduler import Scheduler
from Core.Processor import Processor
from Core.DataPersistent import DataPersistent
from Core.EmailService import EmailService
from Core.Processor import ResponseBuffer
from Core.AccountLogin import AccountManager


class SpiderCore:
    def __init__(self):
        # 默认配置
        # downloader 模块配置
        self.is_proxy_service_enable = False
        self.session_pool_size = 20
        self.download_thread_num = 10
        self.network_retry_times = 3
        self.connect_timeout = 30
        self.download_interval = 3

        # Processor 模块配置
        self.process_thread_num = 2
        self.is_parser_following_list = True
        self.is_parser_follower_list = False
        self.is_parser_follow_relation = False

        # Scheduler 模块配置
        self.url_rate = 8

        # DataPersistent 模块配置
        self.persistent_cache_size = 1000
        self.follow_relation_persistent_cache_size = 1000

        # 邮件服务配置
        self.is_email_service_enable = False
        self.smtp_server_host = ''
        self.smtp_server_port = 25
        self.smtp_server_password = ''
        self.smtp_from_addr = ''
        self.smtp_to_addr = ''
        self.smtp_email_header = ''
        self.smtp_send_interval = 3600

        # redis 数据库配置
        self.redis_host = ''
        self.redis_port = 6379
        self.redis_db = 0
        self.redis_password = ''

        # MySQL 数据库配置
        self.mysql_host = ''
        self.mysql_username = ''
        self.mysql_password = ''
        self.mysql_database = ''
        self.mysql_charset = 'utf8'

        # 知乎账户配置
        self.is_login_by_cookie = True
        self.z_c0 = ''
        self.login_token = ''
        self.password = ''

        # 初始 Token
        self.init_token = []

        # 载入用户自定义配置
        self.load_config()

        # 模块实例
        self.redis_connection = None
        self.mysql_connection = None
        self.response_buffer = None
        self.account_manager = None
        self.downloader = None
        self.processor = None
        self.schedule = None
        self.dataPersistent = None
        self.email_service = None

    # 启动Spider
    def start_spider_core(self):
        if log.isEnabledFor(logging.INFO):
            log.info('Spider 开始启动')

        try:
            # 创建Redis连接
            redis_connect_retry_times = 3
            while redis_connect_retry_times > 0:
                self.redis_connection = redis.StrictRedis(host=self.redis_host,
                                                          port=self.redis_port,
                                                          db=self.redis_db,
                                                          password=self.redis_password)
                ping = self.redis_connection.ping()
                if ping is True:
                    if log.isEnabledFor(logging.INFO):
                        log.info('Redis 服务器连接成功')
                    break
                else:
                    if log.isEnabledFor(logging.INFO):
                        log.info('Redis 服务器连接失败')
                    redis_connect_retry_times -= 1
                    time.sleep(5)

            # 若连接不成功则退出
            if redis_connect_retry_times <= 0:
                raise Exception()

            # 创建MySQL连接
            self.mysql_connection = pymysql.connect(host=self.mysql_host,
                                                    user=self.mysql_username,
                                                    passwd=self.mysql_password,
                                                    db=self.mysql_database,
                                                    charset=self.mysql_charset)

        except Exception as e:
            if log.isEnabledFor(logging.ERROR):
                log.error('Redis 启动失败')
                log.exception(e)
            return

        # 创建 response 缓存队列
        self.response_buffer = ResponseBuffer()

        # 启动账户管理器并登陆
        self.account_manager = AccountManager(self.login_token,
                                              self.password,
                                              self.is_login_by_cookie,
                                              self.z_c0)
        is_login = self.account_manager.login()
        if not is_login:
            return

        # 启动Downloader
        self.downloader = Downloader(self.redis_connection,
                                     self.response_buffer,
                                     self.account_manager,
                                     self.is_proxy_service_enable,
                                     self.session_pool_size,
                                     self.download_thread_num,
                                     self.network_retry_times,
                                     self.connect_timeout,
                                     self.download_interval)
        self.downloader.start_downloader()

        # 启动Scheduler
        self.schedule = Scheduler(self.redis_connection, self.url_rate)
        self.schedule.start()

        # 启动 DataPersistent
        self.dataPersistent = DataPersistent(self.persistent_cache_size,
                                             self.follow_relation_persistent_cache_size,
                                             self.mysql_connection,
                                             self.redis_connection)
        self.dataPersistent.start_data_persistent()

        # 启动Processor
        self.processor = Processor(self.process_thread_num,
                                   self.is_parser_following_list,
                                   self.is_parser_follower_list,
                                   self.is_parser_follow_relation,
                                   self.redis_connection,
                                   self.response_buffer)
        self.processor.start_processor()
        self.processor.load_init_data(self.init_token)

        # 启动邮件服务
        if self.is_email_service_enable is True:
            self.email_service = EmailService(self.smtp_server_host,
                                              self.smtp_server_port,
                                              self.smtp_server_password,
                                              self.smtp_from_addr,
                                              self.smtp_to_addr,
                                              self.smtp_email_header,
                                              self.smtp_send_interval,
                                              self.dataPersistent)
            self.email_service.start_email_service()
            self.email_service.send_message('Spider 启动完毕')

        if log.isEnabledFor(logging.INFO):
            log.info('Spider 启动完毕')

        # 模块异常检查
        while True:
            # Downloader模块异常检查
            self.downloader.check_and_restart()
            # EmailService 模块异常检查
            if self.is_email_service_enable is True:
                self.email_service.check_and_restart()
            # DataPersistent 模块异常检查
            self.dataPersistent.check_and_restart()
            # Scheduler 模块异常检查
            # Processor 模块异常检查
            self.processor.check_and_restart()
            # 检查间隔
            time.sleep(180)
            gc.collect()

    # 加载自定义配置信息
    def load_config(self):
        section = "spider_core"
        config = configparser.ConfigParser()
        config.read("Core/Config/SpiderCoreConfig.conf", encoding="utf8")

        # 读取 downloader 模块配置
        self.is_proxy_service_enable = True if int(config.get(section, 'isProxyServiceEnable')) == 1 else False
        self.session_pool_size = int(config.get(section, 'sessionPoolSize'))
        self.download_thread_num = int(config.get(section, 'downloadThreadNum'))
        self.network_retry_times = int(config.get(section, 'networkRetryTimes'))
        self.connect_timeout = int(config.get(section, 'connectTimeout'))
        self.download_interval = int(config.get(section, 'downloadInterval'))

        # 读取 Processor 模块配置
        self.process_thread_num = int(config.get(section, 'processThreadNum'))
        self.is_parser_following_list = True if int(config.get(section, 'isParserFollowingList')) == 1 else False
        self.is_parser_follower_list = True if int(config.get(section, 'isParserFollowerList')) == 1 else False
        self.is_parser_follow_relation = True if int(config.get(section, 'isParserFollowRelation')) == 1 else False

        # 读取 Scheduler 模块配置
        self.url_rate = int(config.get(section, 'urlRate'))

        # 读取 DataPersistent 模块配置
        self.persistent_cache_size = int(config.get(section, 'persistentCacheSize'))
        self.follow_relation_persistent_cache_size = int(config.get(section, 'followRelationPersistentCacheSize'))

        # 读取邮件服务配置
        self.is_email_service_enable = True if int(config.get(section, 'isEmailServiceEnable')) == 1 else False
        self.smtp_server_host = config.get(section, 'smtpServerHost')
        self.smtp_server_port = int(config.get(section, 'smtpServerPort'))
        self.smtp_server_password = config.get(section, 'smtpServerPassword')
        self.smtp_from_addr = config.get(section, 'smtpFromAddr')
        self.smtp_to_addr = config.get(section, 'smtpToAddr')
        self.smtp_email_header = config.get(section, 'smtpEmailHeader')
        self.smtp_send_interval = int(config.get(section, 'smtpSendInterval'))

        # 读取 Redis 数据库配置
        self.redis_host = config.get(section, 'redisHost')
        self.redis_port = int(config.get(section, 'redisPort'))
        self.redis_db = int(config.get(section, 'redisDB'))
        self.redis_password = config.get(section, 'redisPassword')

        # 读取 MySQL 数据库配置
        self.mysql_host = config.get(section, 'mysqlHost')
        self.mysql_username = config.get(section, 'mysqlUsername')
        self.mysql_password = config.get(section, 'mysqlPassword')
        self.mysql_database = config.get(section, 'mysqlDatabase')
        self.mysql_charset = config.get(section, 'mysqlCharset')

        # 读取知乎账户配置
        self.is_login_by_cookie = True if int(config.get(section, 'isLoginByCookie')) == 1 else False
        self.z_c0 = config.get(section, 'z_c0')
        self.login_token = config.get(section, 'loginToken')
        self.password = config.get(section, 'password')

        # 读取初始token
        token_list = config.get(section, 'initToken')
        for token in token_list.split(','):
            self.init_token.append(str(token).strip())

        if log.isEnabledFor(logging.INFO):
            log.info('配置文件读取并配置完毕')



import threading
import logging
import time
from Core.Logger import log

# 知乎用户信息字段
# 用户头像
USER_AVATAR_URL_TEMPLATE = 'avatarUrlTemplate'
# 用户标识
USER_URL_TOKEN = 'urlToken'
# 用户名
USER_NAME = 'name'
# 用户自我介绍
USER_HEADLINE = 'headline'
# 用户居住地
USER_LOCATIONS = 'locations'
# 用户所在行业
USER_BUSINESS = 'business'
# 用户职业经历
USER_EMPLOYMENTS = 'employments'
# 用户教育经历
USER_EDUCATIONS = 'educations'
# 用户个人描述
USER_DESCRIPTION = 'description'
# 用户性别
USER_GENDER = 'gender'
# 正在关注用户的数目
USER_FOLLOWING_COUNT = 'followingCount'
# 关注者的数目
USER_FOLLOWER_COUNT = 'followerCount'
# 该用户回答问题的数目
USER_ANSWER_COUNT = 'answerCount'
# 用户提问数目
USER_QUESTION_COUNT = 'questionCount'
# 用户获得赞同的数目
USER_VOTE_UP_COUNT = 'voteupCount'

# 关注关系字段
# 关注者
FOLLOW_FROM = 'followFrom'
# 被关注者
FOLLOW_TO = 'followTo'


# 负责用户数据的持久化
class DataPersistent:
    __slots__ = ('persistent_cache_size', 'db_connection', 'redis_connection', 'persistent_thread',
                 'follow_relation_persistent_cache_size')

    # 初始化
    def __init__(self, persistent_cache_size, follow_relation_persistent_cache_size, db_connection, redis_connection):
        # 设置用户信息数据持久化缓存大小
        self.persistent_cache_size = persistent_cache_size
        # 设置用户关注关系持久化缓存大小
        self.follow_relation_persistent_cache_size = follow_relation_persistent_cache_size
        # 设置数据库连接
        self.db_connection = db_connection
        # 设置Redis连接
        self.redis_connection = redis_connection
        # 创建数据库持久化线程
        self.persistent_thread = PersistentThread(self.db_connection,
                                                  self.redis_connection,
                                                  self.persistent_cache_size,
                                                  self.follow_relation_persistent_cache_size)

        if log.isEnabledFor(logging.INFO):
            log.info('DataPersistent 模块初始化完毕')

    def get_current_user_info_num(self):
        return self.persistent_thread.get_current_user_info_num()

    # 启动DataPersistent模块
    def start_data_persistent(self):
        # 启动线程
        self.persistent_thread.start()

        if log.isEnabledFor(logging.INFO):
            log.info('DataPersistent 模块启动成功')

        # 线程异常检测并重启
    def check_and_restart(self):
        if self.persistent_thread.thread_status == 'error':
            self.persistent_thread = PersistentThread(self.db_connection, self.redis_connection,
                                                      self.persistent_cache_size,
                                                      self.follow_relation_persistent_cache_size)
            self.persistent_thread.start()
            if log.isEnabledFor(logging.INFO):
                log.info('DataPersistent模块持久化线程中重新启动')


# 用户信息数据插入SQL语句
INSERT_USER_INFO = 'insert ignore into user_info(user_avator_url, user_token, user_name, user_headline, ' \
                   'user_location, user_business, user_employments, user_educations, user_description, ' \
                   'user_gender, user_following_count, user_follower_count, user_answer_count, ' \
                   'user_question_count, user_voteup_count) values(%s,%s,%s,%s,%s,%s,%s,%s,' \
                   '%s,%s,%s,%s,%s,%s,%s)'

# 用户关注关系数据插入SQL语句
INSERT_FOLLOW_RELATION = 'insert ignore into follow_relation(follow_from, follow_in) values(%s, %s)'

# 用户信息数量查询语句
COUNT_USER_INFO = 'select count(*) from user_info'


class PersistentThread(threading.Thread):
    def __init__(self, db_connection, redis_connection, persistent_cache_size, follow_relation_persistent_cache_size):
        threading.Thread.__init__(self)
        # 设置数据库连接
        self.db_connection = db_connection
        # 设置Redis连接
        self.redis_connection = redis_connection
        # 设置用户信息缓存大小
        self.persistent_cache_size = persistent_cache_size
        # 设置关注关系缓存大小
        self.follow_relation_persistent_cache_size = follow_relation_persistent_cache_size
        # 数据持久化缓存队列名称
        self.persistent_cache = 'persistentCache'
        # 关注关系持久化缓存队列名称
        self.follow_relation_persistent_cache = 'followRelationPersistentCache'
        # 线程状态
        self.thread_status = 'working'
        # Operation Lock
        self.lock = threading.Lock()

    def get_current_user_info_num(self):
        self.lock.acquire()
        # 获取存放在数据库中的用户数量
        cursor = self.db_connection.cursor()
        cursor.execute(COUNT_USER_INFO)
        data = cursor.fetchone()
        cursor.close()
        user_num_db = data[0]

        # 获取缓存队列中的用户数量
        user_num_cache = self.redis_connection.llen(self.persistent_cache)
        self.lock.release()
        return user_num_cache + user_num_db

    def run(self):
        debug_info = None
        try:
            while True:
                # 持久化用户信息
                current_user_info_cache_size = self.redis_connection.llen(self.persistent_cache)
                if current_user_info_cache_size >= self.persistent_cache_size:
                    self.lock.acquire()
                    cursor = self.db_connection.cursor()
                    for i in range(current_user_info_cache_size):
                        user_info = self.redis_connection.lpop(self.persistent_cache)
                        debug_info = user_info
                        if user_info is not None:
                            user_info = self.convert_user_info(eval(user_info.decode('utf-8')))
                            cursor.execute(INSERT_USER_INFO, [user_info[USER_AVATAR_URL_TEMPLATE],
                                                              user_info[USER_URL_TOKEN],
                                                              user_info[USER_NAME],
                                                              user_info[USER_HEADLINE],
                                                              user_info[USER_LOCATIONS],
                                                              user_info[USER_BUSINESS],
                                                              user_info[USER_EMPLOYMENTS],
                                                              user_info[USER_EDUCATIONS],
                                                              user_info[USER_DESCRIPTION],
                                                              user_info[USER_GENDER],
                                                              user_info[USER_FOLLOWING_COUNT],
                                                              user_info[USER_FOLLOWER_COUNT],
                                                              user_info[USER_ANSWER_COUNT],
                                                              user_info[USER_QUESTION_COUNT],
                                                              user_info[USER_VOTE_UP_COUNT]])
                    self.db_connection.commit()
                    cursor.close()
                    self.lock.release()

                # 持久化关注关系
                current_follow_relation_cache_size = self.redis_connection.llen(self.follow_relation_persistent_cache)
                if current_follow_relation_cache_size >= self.follow_relation_persistent_cache_size:
                    self.lock.acquire()
                    cursor = self.db_connection.cursor()
                    for i in range(current_follow_relation_cache_size):
                        follow_relation = self.redis_connection.lpop(self.follow_relation_persistent_cache)
                        debug_info = follow_relation
                        if follow_relation is not None:
                            follow_relation = eval(follow_relation.decode('utf-8'))
                            cursor.execute(INSERT_FOLLOW_RELATION, [follow_relation[FOLLOW_FROM],
                                                                    follow_relation[FOLLOW_TO]])
                    self.db_connection.commit()
                    cursor.close()
                    self.lock.release()

                # 检查时间间隔
                time.sleep(180)

        except Exception as e:
            if log.isEnabledFor(logging.ERROR):
                log.error('用户数据持久化线程异常退出')
                log.exception(e)
            log.debug(debug_info)
            self.thread_status = 'error'

    # 将用户数据转换为适合数据库存储的格式
    @staticmethod
    def convert_user_info(user_info):
        # 将居住地转换为‘；’分隔的字符串
        locations_string = ';'.join(str(x) for x in user_info[USER_LOCATIONS])
        user_info[USER_LOCATIONS] = locations_string

        # 将职业经历转换为‘XXX（XXX）’，并以‘；’ 分隔的字符串
        employments_list = []
        for employment in user_info[USER_EMPLOYMENTS]:
            temp = ''
            if 'company' in employment:
                temp += str(employment['company'])
            if 'job' in employment:
                temp += '-' + str(employment['job'])
            employments_list.append(temp)
        employments_string = ';'.join(str(x) for x in employments_list)
        user_info[USER_EMPLOYMENTS] = employments_string

        # 将教育经历转换为‘；’分隔的字符串
        educations_string = ';'.join(str(x) for x in user_info[USER_EDUCATIONS])
        user_info[USER_EDUCATIONS] = educations_string

        return user_info



import datetime
import threading
import smtplib
import logging
import time
from email.header import Header
from email.mime.text import MIMEText
from Core.Logger import log


class EmailService:
    __slots__ = ('data_persistent', 'smtp_server_host', 'smtp_server_port', 'smtp_server_password',
                 'smtp_from_addr', 'smtp_to_addr', 'smtp_email_header', 'smtp_send_interval', 'email_service_thread')

    def __init__(self, smtp_server_host, smtp_server_port, smtp_server_password,
                 smtp_from_addr, smtp_to_addr, smtp_email_header, smtp_send_interval, data_persistent):
        self.data_persistent = data_persistent
        # 设置参数
        self.smtp_server_host = smtp_server_host
        self.smtp_server_port = smtp_server_port
        self.smtp_server_password = smtp_server_password
        self.smtp_from_addr = smtp_from_addr
        self.smtp_to_addr = smtp_to_addr
        self.smtp_email_header = smtp_email_header
        self.smtp_send_interval = smtp_send_interval

        # 创建邮件定时发送线程
        self.email_service_thread = EmailServiceThread(self.smtp_server_host,
                                                       self.smtp_server_port,
                                                       self.smtp_server_password,
                                                       self.smtp_from_addr,
                                                       self.smtp_to_addr,
                                                       self.smtp_email_header,
                                                       self.smtp_send_interval,
                                                       self.data_persistent)

        if log.isEnabledFor(logging.INFO):
            log.info('EmailService 模块初始化完毕')

    # 启动邮件定时发送线程
    def start_email_service(self):
        # 启动线程
        self.email_service_thread.start()

        if log.isEnabledFor(logging.INFO):
            log.info('EmailService 模块启动成功')

    def check_and_restart(self):
        if self.email_service_thread.thread_status == 'error':
            self.email_service_thread = EmailServiceThread(self.smtp_server_host,
                                                           self.smtp_server_port,
                                                           self.smtp_server_password,
                                                           self.smtp_from_addr,
                                                           self.smtp_to_addr,
                                                           self.smtp_email_header,
                                                           self.smtp_send_interval,
                                                           self.data_persistent)
            self.email_service_thread.start()
            if log.isEnabledFor(logging.INFO):
                log.info('EmailService线程重新启动')

    # 发送指定的内容
    def send_message(self, email_content):
        # 准备发送的内容
        now = datetime.datetime.now()
        header = self.smtp_email_header + '[' + str(now.month) + '-' + str(now.day) + ' ' + \
            str(now.hour) + ':' + str(now.minute) + ':' + str(now.second) + ']'
        msg = MIMEText(email_content, 'plain', 'utf-8')
        msg['from'] = self.smtp_from_addr
        msg['to'] = self.smtp_to_addr
        msg['Subject'] = Header(header, 'utf-8').encode()

        # 发送
        try:
            smtp_server = smtplib.SMTP(self.smtp_server_host, self.smtp_server_port)
            smtp_server.login(self.smtp_from_addr, self.smtp_server_password)
            smtp_server.sendmail(self.smtp_from_addr, [self.smtp_to_addr], msg.as_string())
            smtp_server.quit()
        except Exception as e:
            if log.isEnabledFor(logging.ERROR):
                log.error("邮件发送失败")
                log.exception(e)


class EmailServiceThread(threading.Thread):
    def __init__(self, smtp_server_host, smtp_server_port, smtp_server_password,
                 smtp_from_addr, smtp_to_addr, smtp_email_header, smtp_send_interval, data_persistent):
        threading.Thread.__init__(self)
        self.data_persistent = data_persistent
        # 设置参数
        self.smtp_server_host = smtp_server_host
        self.smtp_server_port = smtp_server_port
        self.smtp_server_password = smtp_server_password
        self.smtp_from_addr = smtp_from_addr
        self.smtp_to_addr = smtp_to_addr
        self.smtp_email_header = smtp_email_header
        self.smtp_send_interval = smtp_send_interval
        # 线程状态
        self.thread_status = 'working'
        # 持久化队列名称
        self.persistent_cache = 'persistentCache'

        self.lastSendTime = datetime.datetime.now()
        self.lastUserInfoNum = self.data_persistent.get_current_user_info_num()

    def run(self):
        if log.isEnabledFor(logging.DEBUG):
            log.debug("邮件服务线程启动")
        try:
            while True:
                time.sleep(self.smtp_send_interval)

                # 准备发送的内容
                msg = MIMEText(self.get_email_content(), 'plain', 'utf-8')
                msg['from'] = self.smtp_from_addr
                msg['to'] = self.smtp_to_addr
                msg['Subject'] = Header(self.get_email_header(), 'utf-8').encode()

                # 发送
                smtp_server = smtplib.SMTP(self.smtp_server_host, self.smtp_server_port)
                smtp_server.login(self.smtp_from_addr, self.smtp_server_password)
                smtp_server.sendmail(self.smtp_from_addr, [self.smtp_to_addr], msg.as_string())
                smtp_server.quit()

                # 更新最后一次发送时间
                self.lastSendTime = datetime.datetime.now()

        except Exception as e:
            if log.isEnabledFor(logging.ERROR):
                log.error("邮件发送失败")
                log.exception(e)
            self.thread_status = 'error'

    # 构造邮件标题
    def get_email_header(self):
        now = datetime.datetime.now()
        return self.smtp_email_header + '[' + str(now.month) + '-' + str(now.day) + ' ' \
            + str(now.hour) + ':' + str(now.minute) + ':' + str(now.second) + ']'

    # 构建邮件内容
    def get_email_content(self):
        # 获得当前用户的数量
        current_num = self.data_persistent.get_current_user_info_num()
        total_add = current_num - self.lastUserInfoNum
        self.lastUserInfoNum = current_num

        # 构建邮件内容
        info_line = '自' + '[' + str(self.lastSendTime.month) + '-' + str(self.lastSendTime.day) + ' ' \
                    + str(self.lastSendTime.hour) + ':' + str(self.lastSendTime.minute) \
                    + ':' + str(self.lastSendTime.second) + ']' + "以来,\n" + "爬取到新的用户数量为：" + str(total_add) \
                    + '\n当前爬取到的用户信息总条数为：' + str(current_num) + '\n\n\n\n'

        # 获取日志信息
        logging_file = open('Logs/ZhiHuSpider.log', encoding='utf8')
        logging_info_line = '日志文件信息：\n' + logging_file.read()
        logging_file.close()

        return info_line + logging_info_line



from Proxy import proxyCore
from Core.Logger import log
import queue
import threading
import logging
import time
import requests

logging.getLogger("requests").setLevel(logging.WARNING)
logging.getLogger("urllib3").setLevel(logging.WARNING)

# 请求头信息
requestHeader = {"Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,*/*;q=0.8",
                 "Accept-Encoding": "gzip, deflate, sdch, br",
                 "Accept-Language": "zh-CN,zh;q=0.8",
                 "Cache-Control": "max-age=0",
                 "Host": "www.zhihu.com",
                 "Upgrade-Insecure-Requests": "1",
                 "User-Agent": "Mozilla/5.0 (Windows NT 10.0; WOW64) AppleWebKit/537.36 (KHTML, like Gecko)"
                               " Chrome/56.0.2924.87 Safari/537.36"}


# SessionPool 管理器
class SessionManager:
    __slots__ = ('session_pool_size', 'created_session_num', 'available_session_num', 'is_proxy_service_enable',
                 'account_manager', 'session_pool', 'proxy_service', 'available_session_lock', 'created_session_lock')

    # 初始化
    def __init__(self, session_pool_size, account_manager, is_proxy_service_enable):
        # session pool 大小
        self.session_pool_size = session_pool_size
        # 已经创建的session数量
        self.created_session_num = 0
        # 当前池中的session数量
        self.available_session_num = 0
        # 是否启用代理服务
        self.is_proxy_service_enable = is_proxy_service_enable
        # 账号认证管理器
        self.account_manager = account_manager

        # available session num 锁
        self.available_session_lock = threading.Lock()
        # created session num 锁
        self.created_session_lock = threading.Lock()

        # 创建 session pool
        self.session_pool = queue.Queue(session_pool_size)
        # 创建并启动代理服务
        if self.is_proxy_service_enable is True:
            self.proxy_service = proxyCore.ProxyService()
            self.proxy_service.start_proxy_service()

        if log.isEnabledFor(logging.INFO):
            log.info("Session Manager 启动成功")

    # 创建包含代理信息的 session
    def create_session_proxy(self, proxy_info):
        # 获取代理信息
        ip = proxy_info['ip']
        port = proxy_info['port']
        protocol = proxy_info['protocol'].lower()
        proxy = {protocol: ip + ':' + port}

        # 创建session
        session = requests.session()
        session.headers = requestHeader
        session.cookies.update(self.account_manager.get_auth_token())
        session.proxies = proxy

        # 将 session 放入到池中
        self.session_pool.put(session)
        self.created_session_num_change(1)
        self.available_session_num_change(1)

    # 创建不含代理信息的 session
    def create_session(self):
        # 创建session
        session = requests.session()
        session.cookies.update(self.account_manager.get_auth_token())
        session.headers = requestHeader

        # 将 session 放入池中
        self.session_pool.put(session)
        self.created_session_num_change(1)
        self.available_session_num_change(1)

    # 获取一个session连接
    def get_session_connection(self):

        # 辅助创建 session pool 中的 session
        if self.available_session_num <= 0 and self.created_session_num < self.session_pool_size:
            if self.is_proxy_service_enable is True:
                self.create_session_proxy(self.proxy_service.get_proxy())
            else:
                self.create_session()

        # 从 session pool 中获取
        session = self.session_pool.get()
        self.available_session_num_change(-1)
        return session

    # 归还一个 session 连接
    def return_session_connection(self, session_connection):
        self.session_pool.put(session_connection)
        self.available_session_num_change(1)

    # 归还并更换 session 的代理
    def return_and_switch_proxy(self, session_connection):
        # 更换代理信息
        if self.is_proxy_service_enable is True:
            proxy_info = self.proxy_service.get_proxy()
            ip = proxy_info['ip']
            port = proxy_info['port']
            protocol = proxy_info['protocol'].lower()
            proxy = {protocol: ip + ':' + port}
            session_connection.proxies = proxy
        # 归还
        self.return_session_connection(session_connection)

    # 修改 available_session_num 的值
    def available_session_num_change(self, num):
        self.available_session_lock.acquire()
        self.available_session_num += num
        self.available_session_lock.release()

    # 修改 available_session_num 的值
    def created_session_num_change(self, num):
        self.created_session_lock.acquire()
        self.created_session_num += num
        self.created_session_lock.release()


# 管理下载线程
class Downloader:
    __slots__ = ('download_thread_num', 'redis_connection', 'response_buffer', 'session_manager',
                 'NETWORK_RETRY_TIMES', 'account_manager',
                 'CONNECT_TIMEOUT', 'DOWNLOAD_INTERVAL', 'download_thread_list')

    # 初始化
    def __init__(self, redis_connection, response_buffer, account_manager, is_proxy_service_enable, session_pool_size,
                 download_thread_num, network_retry_times, connect_timeout, download_interval):
        # 设置下载线程的数量
        self.download_thread_num = download_thread_num
        # 设置 Redis 连接
        self.redis_connection = redis_connection
        # 设置 response 缓存队列
        self.response_buffer = response_buffer
        # 设置账户认证管理器
        self.account_manager = account_manager
        # 设置并启动sessionManager
        self.session_manager = SessionManager(session_pool_size, account_manager, is_proxy_service_enable)

        # 设置网络连接参数
        self.NETWORK_RETRY_TIMES = network_retry_times
        self.CONNECT_TIMEOUT = connect_timeout
        self.DOWNLOAD_INTERVAL = download_interval

        # 初始化下载线程
        self.download_thread_list = []
        for i in range(self.download_thread_num):
            download_thread = DownloadThread('thread' + str(i), self.session_manager, self.redis_connection,
                                             self.response_buffer, self.NETWORK_RETRY_TIMES, self.CONNECT_TIMEOUT,
                                             self.DOWNLOAD_INTERVAL)
            self.download_thread_list.append(download_thread)

        if log.isEnabledFor(logging.INFO):
            log.info("Downloader 模块初始化完毕")

    # 启动
    def start_downloader(self):
        # 启动下载线程
        for download_thread in self.download_thread_list:
            download_thread.start()

        if log.isEnabledFor(logging.INFO):
            log.info('Downloader 模块启动成功')

    # 检查并重启线程
    def check_and_restart(self):
        for download_thread in self.download_thread_list:
            if download_thread.thread_status == 'error':
                thread_id = download_thread.thread_id
                self.download_thread_list.remove(download_thread)
                download_thread = DownloadThread(thread_id, self.session_manager, self.redis_connection,
                                                 self.response_buffer, self.NETWORK_RETRY_TIMES, self.CONNECT_TIMEOUT,
                                                 self.DOWNLOAD_INTERVAL)
                self.download_thread_list.append(download_thread)
                download_thread.start()
                if log.isEnabledFor(logging.INFO):
                    log.info('数据下载线程' + thread_id + '重新启动')


# 下载线程
class DownloadThread(threading.Thread):
    __slots__ = ('thread_status', 'thread_id', 'session_manager', 'redis_connection', 'response_buffer',
                 'url_queue_name', 'response_queue_name', 'NETWORK_RETRY_TIMES', 'CONNECT_TIMEOUT',
                 'DOWNLOAD_INTERVAL')

    def __init__(self, thread_id, session_manager, redis_connection, response_buffer, network_retry_times,
                 connect_timeout, download_interval):
        threading.Thread.__init__(self)
        # 线程状态
        self.thread_status = 'working'
        # 线程ID
        self.thread_id = thread_id
        # 设置session manager
        self.session_manager = session_manager
        # 设置 redis 连接
        self.redis_connection = redis_connection
        # 设置 response 缓存队列
        self.response_buffer = response_buffer
        # 设置获取待下载URL队列名称
        self.url_queue_name = 'urlQueue'
        # 设置存放下载内容队列名称
        self.response_queue_name = 'responseQueue'

        # 设置网络连接的参数
        self.NETWORK_RETRY_TIMES = network_retry_times
        self.CONNECT_TIMEOUT = connect_timeout
        self.DOWNLOAD_INTERVAL = download_interval

    def run(self):
        if log.isEnabledFor(logging.INFO):
            log.info('数据下载线程' + self.thread_id + '启动')

        # 保存上一次未下载的url info
        previous_url_info = None
        while True:
            # 获取session
            session = self.session_manager.get_session_connection()

            # 尝试下载数据
            network_retry_times = 0
            while network_retry_times < self.NETWORK_RETRY_TIMES:
                try:
                    # 获取URL
                    if previous_url_info is None:
                        url_info = self.get_url_info_from_queue()
                        previous_url_info = url_info
                    else:
                        url_info = previous_url_info
                    url = url_info[1]

                    # 下载数据
                    response = session.get(url, timeout=self.CONNECT_TIMEOUT)

                    if log.isEnabledFor(logging.DEBUG):
                        log.debug(response.status_code)

                    # 检查返回结果
                    if response.status_code == 200:
                        # 封装下载的数据(包括原来的数据)
                        response_info = url_info
                        response_info[1] = response.text
                        self.put_response_info_to_queue(response_info)
                        previous_url_info = None
                        if log.isEnabledFor(logging.DEBUG):
                            log.debug('下载成功')
                        break
                    elif response.status_code == 403:
                        if log.isEnabledFor(logging.ERROR):
                            log.error('账号认证失败')
                        break
                    elif response.status_code == 429:
                        if log.isEnabledFor(logging.DEBUG):
                            log.debug('[' + str(self.thread_id) + ']' + '访问太频繁，稍候重新访问，响应码为：'
                                      + str(response.status_code))
                        previous_url_info = url_info
                        break
                    elif response.status_code == 404 or response.status_code == 410:
                        previous_url_info = None
                        break
                    else:
                        if log.isEnabledFor(logging.ERROR):
                            log.error(response.status_code)
                        network_retry_times += 1
                except Exception as e:
                    network_retry_times += 1
                    time.sleep(self.DOWNLOAD_INTERVAL)
                    if log.isEnabledFor(logging.DEBUG):
                        log.debug('[' + str(self.thread_id) + ']' + '下载异常，正在重新连接...（第'
                                  + str(network_retry_times) + '次重试)')
                    if log.isEnabledFor(logging.DEBUG):
                        log.error(e)

            # 下载间隔
            time.sleep(self.DOWNLOAD_INTERVAL)

            # 归还session
            if network_retry_times < self.NETWORK_RETRY_TIMES:
                self.session_manager.return_session_connection(session)
            else:
                self.session_manager.return_and_switch_proxy(session)

    # 从队列中获取一个待下载的URL
    def get_url_info_from_queue(self):
        while True:
            url_info = self.redis_connection.blpop(self.url_queue_name, 10)
            if url_info is not None and len(url_info) >= 2:
                url_info = eval(url_info[1].decode('utf-8'))
                if url_info is not None:
                    return url_info

    # 将下载的网页内容放入到队列中
    def put_response_info_to_queue(self, response):
        self.response_buffer.put_response_to_buffer(response)
        del response



from bs4 import BeautifulSoup


class DataParseModule:
    def __init__(self):
        pass

    # 解析 DOM 内容
    @staticmethod
    def parse_data(data):
        from Proxy import proxyCore
        # 存放解析出来代理信息
        proxy_ip_list = []

        # 判断数据是否为空
        if data is None:
            return proxy_ip_list

        # 解析获得包含代理 IP 的表格
        html = BeautifulSoup(data, "html.parser")
        html_ip_list = html.find('table', {'id': 'ip_list'})
        if html_ip_list is None:
            return proxy_ip_list
        data_list = html_ip_list.findAll('tr')

        # 逐个解析表格中的项
        for elem in data_list:
            if elem.find('th', text="国家") is not None:
                continue

            # 解析代理 IP 信息
            country_node = elem.find('td')
            ip_node = country_node.find_next_sibling()
            port_node = ip_node.find_next_sibling()
            protocol_node = port_node.find_next_sibling().find_next_sibling().find_next_sibling()
            ip = ip_node.string
            port = port_node.string
            protocol = protocol_node.string

            # 封装代理信息
            proxy_ip_info = {proxyCore.PROXY_IP: ip,
                             proxyCore.PROXY_PORT: port,
                             proxyCore.PROXY_PROTOCOL: protocol
                             }

            proxy_ip_list.append(proxy_ip_info)
        return proxy_ip_list



import threading
import logging
import html
import json
import time
import queue
from Core.Logger import log
from bs4 import BeautifulSoup

# 知乎用户信息字段
# 用户头像
USER_AVATAR_URL_TEMPLATE = 'avatarUrlTemplate'
# 用户标识
USER_URL_TOKEN = 'urlToken'
# 用户名
USER_NAME = 'name'
# 用户自我介绍
USER_HEADLINE = 'headline'
# 用户居住地
USER_LOCATIONS = 'locations'
# 用户所在行业
USER_BUSINESS = 'business'
# 用户职业经历
USER_EMPLOYMENTS = 'employments'
# 用户教育经历
USER_EDUCATIONS = 'educations'
# 用户个人描述
USER_DESCRIPTION = 'description'
# 用户性别
USER_GENDER = 'gender'
# 正在关注用户的数目
USER_FOLLOWING_COUNT = 'followingCount'
# 关注者的数目
USER_FOLLOWER_COUNT = 'followerCount'
# 该用户回答问题的数目
USER_ANSWER_COUNT = 'answerCount'
# 用户提问数目
USER_QUESTION_COUNT = 'questionCount'
# 用户获得赞同的数目
USER_VOTE_UP_COUNT = 'voteupCount'

# 关注关系字段
# 关注者
FOLLOW_FROM = 'followFrom'
# 被关注者
FOLLOW_TO = 'followTo'


# 下载数据处理器
class Processor:
    __slots__ = ('process_thread_num', 'redis_connection', 'token_filter', 'response_buffer',
                 'is_parser_following_list', 'is_parser_follower_list', 'processor_list', 'is_parser_follow_relation')

    # 初始化
    def __init__(self, process_thread_num, is_parser_following_list, is_parser_follower_list, is_parser_follow_relation,
                 redis_connection, response_buffer):
        # 设置数据处理器数量
        self.process_thread_num = process_thread_num
        # 设置 Redis 连接
        self.redis_connection = redis_connection
        # 创建 Token 过滤器
        self.token_filter = TokenFilter(self.redis_connection)
        # 设置 response 缓存队列
        self.response_buffer = response_buffer

        # 是否解析正在关注列表
        self.is_parser_following_list = is_parser_following_list
        # 是否解析关注者列表
        self.is_parser_follower_list = is_parser_follower_list
        # 是否解析关注关系
        self.is_parser_follow_relation = is_parser_follow_relation

        # 创建处理器
        self.processor_list = []
        for i in range(process_thread_num):
            process_thread = ProcessThread('thread' + str(i), self.redis_connection, self.token_filter,
                                           self.response_buffer, self.is_parser_following_list,
                                           self.is_parser_follower_list, self.is_parser_follow_relation)
            self.processor_list.append(process_thread)

        if log.isEnabledFor(logging.INFO):
            log.info('Processor 模块初始化完毕')

    # 载入初始 token
    def load_init_data(self, token_list):
        if token_list is None:
            return

        for token in token_list:
            # 封装 URL 信息
            url_info = ['info', URL_PUBLIC + token + URL_PINS, token]
            self.redis_connection.rpush('userInfoURLQueue', url_info)
            del url_info

        if log.isEnabledFor(logging.INFO):
            log.info('初始用户Token载入完毕')

    # 启动数据处理器
    def start_processor(self):
        # 启动处理线程
        for process_thread in self.processor_list:
            process_thread.start()

        if log.isEnabledFor(logging.INFO):
            log.info('Processor 模块启动成功')

    # 异常检查
    def check_and_restart(self):
        for process_thread in self.processor_list:
            if process_thread.thread_status == 'error':
                thread_id = process_thread.thread_id
                self.processor_list.remove(process_thread)
                del process_thread
                new_thread = ProcessThread(thread_id, self.redis_connection, self.token_filter,
                                           self.response_buffer, self.is_parser_following_list,
                                           self.is_parser_follower_list, self.is_parser_follow_relation)
                self.processor_list.append(new_thread)
                new_thread.start()

                if log.isEnabledFor(logging.ERROR):
                    log.error('数据处理器线程[' + thread_id + ']重新启动')

# URL 组件
URL_PUBLIC = 'https://www.zhihu.com/people/'
URL_PINS = '/pins'
str1 = 'https://www.zhihu.com/api/v4/members/'
str2 = '/followees?offset='
str3 = '/followers?offset='
str4 = '&limit='


# 处理线程
class ProcessThread(threading.Thread):
    __slots__ = ('thread_id', 'thread_status', 'redis_connection', 'token_filter', 'response_buffer',
                 'user_info_url_queue', 'follow_info_url_queue', 'persistent_cache', 'is_parser_following_list',
                 'is_parser_follower_list', 'follow_relation_persistent_cache', 'is_parser_follow_relation')

    def __init__(self, thread_id, redis_connection, token_filter, response_buffer, is_parser_following_list,
                 is_parser_follower_list, is_parser_follow_relation):
        threading.Thread.__init__(self)
        # 设置线程名称
        self.thread_id = thread_id
        # 设置线程状态
        self.thread_status = 'working'
        # 设置 Redis 连接
        self.redis_connection = redis_connection
        # 设置 token 过滤器
        self.token_filter = token_filter

        # 是否解析正在关注列表
        self.is_parser_following_list = is_parser_following_list
        # 是否解析关注者列表
        self.is_parser_follower_list = is_parser_follower_list
        # 是否解析关注列表
        self.is_parser_follow_relation = is_parser_follow_relation

        # 下载数据队列
        self.response_buffer = response_buffer
        # user info url 队列名称
        self.user_info_url_queue = 'userInfoURLQueue'
        # follower & following list url 队列名称
        self.follow_info_url_queue = 'followInfoURLQueue'
        # 用户信息数据数据持久化缓存队列名称
        self.persistent_cache = 'persistentCache'
        # 关注关系持久化缓存队列名称
        self.follow_relation_persistent_cache = 'followRelationPersistentCache'

    def run(self):
        if log.isEnabledFor(logging.INFO):
            log.info('数据处理线程' + self.thread_id + '启动')

        try:
            while True:
                # 获取Response数据
                response = self.response_buffer.get_response_from_buffer()

                if response is None or len(response) < 2:
                    continue

                # 判断Response类型
                response_type = response[0]

                # 分派给对应的处理方法
                if response_type == 'info':
                    self.parse_user_info(response)
                elif response_type == 'list':
                    self.parse_follow_info(response)
                time.sleep(0.1)

        except Exception as e:
            self.thread_status = 'error'
            if log.isEnabledFor(logging.ERROR):
                log.exception(e)

    # 解析用户信息(生成的Follow URL Info内容[list, url, token, followingList/followerList])
    def parse_user_info(self, response_info):
        # 获取ResponseInfo中的信息
        data = response_info[1]
        token = response_info[2]

        # 提取JSON信息
        user_info_entities = None
        try:
            bs_obj = BeautifulSoup(data, 'html.parser')
            data_json = bs_obj.find('div', attrs={'id': 'data'})
            if data_json is None:
                return
            else:
                data_json = data_json['data-state']

            # 字符串处理
            # 处理转义字符
            data_json = html.unescape(data_json)
            # 处理html标签
            data_json = BeautifulSoup(data_json, 'html.parser').text

            # 转换为JSON对象
            data_json = json.loads(data_json)

            # 提取实体
            if 'entities' not in data_json:
                return
            entities = data_json['entities']

            # 提取用户信息
            if 'users' not in entities:
                return
            users = entities['users']

            # 提取目标用户信息
            if token not in users:
                return
            user_info = users[token]

            # 提取目标用户的个人信息
            avatar_url_template = None
            name = None
            headline = None
            locations = []
            business = None
            employments = []
            educations = []
            description = None
            gender = None
            following_count = None
            follower_count = None
            answer_count = None
            question_count = None
            voteup_count = None
            if USER_AVATAR_URL_TEMPLATE in user_info:
                avatar_url_template = user_info[USER_AVATAR_URL_TEMPLATE]

            if USER_NAME in user_info:
                name = user_info[USER_NAME]

            if USER_HEADLINE in user_info:
                headline = user_info[USER_HEADLINE]

            if USER_LOCATIONS in user_info:
                for location in user_info[USER_LOCATIONS]:
                    locations.append(location['name'])

            if USER_BUSINESS in user_info:
                business = user_info[USER_BUSINESS]['name']

            if USER_EMPLOYMENTS in user_info:
                for employment in user_info[USER_EMPLOYMENTS]:
                    elem = {}
                    if 'job' in employment:
                        job = employment['job']['name']
                        elem.update({'job': job})
                    if 'company' in employment:
                        company = employment['company']['name']
                        elem.update({'company': company})
                    employments.append(elem)

            if USER_EDUCATIONS in user_info:
                for education in user_info[USER_EDUCATIONS]:
                    if 'school' in education:
                        school = education['school']['name']
                        educations.append(school)

            if USER_DESCRIPTION in user_info:
                description = user_info[USER_DESCRIPTION]

            if USER_GENDER in user_info:
                gender = user_info[USER_GENDER]

            if USER_FOLLOWING_COUNT in user_info:
                following_count = user_info[USER_FOLLOWING_COUNT]

            if USER_FOLLOWER_COUNT in user_info:
                follower_count = user_info[USER_FOLLOWER_COUNT]

            if USER_ANSWER_COUNT in user_info:
                answer_count = user_info[USER_ANSWER_COUNT]

            if USER_QUESTION_COUNT in user_info:
                question_count = user_info[USER_QUESTION_COUNT]

            if USER_VOTE_UP_COUNT in user_info:
                voteup_count = user_info[USER_VOTE_UP_COUNT]

            # 构造用户信息实体
            user_info_entities = {USER_AVATAR_URL_TEMPLATE: avatar_url_template,
                                  USER_URL_TOKEN: token,
                                  USER_NAME: name,
                                  USER_HEADLINE: headline,
                                  USER_LOCATIONS: locations,
                                  USER_BUSINESS: business,
                                  USER_EMPLOYMENTS: employments,
                                  USER_EDUCATIONS: educations,
                                  USER_DESCRIPTION: description,
                                  USER_GENDER: gender,
                                  USER_FOLLOWING_COUNT: following_count,
                                  USER_FOLLOWER_COUNT: follower_count,
                                  USER_ANSWER_COUNT: answer_count,
                                  USER_QUESTION_COUNT: question_count,
                                  USER_VOTE_UP_COUNT: voteup_count}

        except Exception as e:
            if log.isEnabledFor(logging.ERROR):
                log.error('User info 数据解析错误')
                log.exception(e)

        # 处理提取的信息
        if user_info_entities is None:
            return

        # 再次检查用户是否已经添加,若已经添加则不再继续
        if self.token_filter.check_token(token) is True:
            return

        # 标记提取的用户信息
        self.token_filter.mark_token(token)

        # 生成 Following List URL
        if self.is_parser_following_list is True:
            pipe = self.redis_connection.pipeline()
            following_count = user_info_entities[USER_FOLLOWING_COUNT]
            if following_count is not None:
                offset = 0
                limit = 20
                while offset < following_count:
                    url_info = ['list', self.generate_following_info_url(token, offset, limit), token, 'followingList']
                    offset += limit
                    pipe.rpush(self.follow_info_url_queue, url_info)
                pipe.execute()

        # 生成 Follower List URL
        if self.is_parser_follower_list is True:
            pipe = self.redis_connection.pipeline()
            follower_count = user_info_entities[USER_FOLLOWER_COUNT]
            if follower_count is not None:
                offset = 0
                limit = 20
                while offset < follower_count:
                    url_info = ['list', self.generate_follower_info_url(token, offset, limit), token, 'followerList']
                    offset += limit
                    pipe.rpush(self.follow_info_url_queue, url_info)
                pipe.execute()

        # 保存提取到的用户信息
        if log.isEnabledFor(logging.DEBUG):
            log.info('成功获取一个用户的详细信息')
        self.redis_connection.rpush(self.persistent_cache, user_info_entities)

    # 解析follower & following 信息(生成的User URLInfo内容格式[info, url, token])
    def parse_follow_info(self, response_info):
        # 获取ResponseInfo中的信息
        data = response_info[1]

        # 提取JSON
        follow_list_token = []
        try:
            # 转为JSON 对象
            json_data = json.loads(data)

            # 提取用户列表信息
            if 'data' not in json_data:
                return
            data = json_data['data']

            # 提取用户 token
            for follow_info in data:
                if 'url_token' in follow_info:
                    token = follow_info['url_token']
                    # 检查重复并添加
                    if self.token_filter.check_token(token) is False:
                        follow_list_token.append(token)

        except Exception as e:
            if log.isEnabledFor(logging.ERROR):
                log.error('Follower & Following List 数据解析错误')
            if log.isEnabledFor(logging.DEBUG):
                log.exception(e)
            return

        # 添加token url 到队列中
        for token in follow_list_token:
            # 封装 URL 信息(List)
            url_info = ['info', self.generate_user_info_url(token), token]
            self.redis_connection.rpush(self.user_info_url_queue, url_info)

        # 提取用户的关注关系(即 following)
        # (返回的Response内容[info, data, token, followingList/followerList])
        if self.is_parser_follow_relation is True:
            # 关注列表类型
            follow_list_type = response_info[3]
            # 用户Token
            token = response_info[2]
            if follow_list_type == 'followingList':
                pipe = self.redis_connection.pipeline()
                for following_token in follow_list_token:
                    # 封装关注关系
                    follow_relation = {FOLLOW_FROM: token,
                                       FOLLOW_TO: following_token}
                    pipe.rpush(self.follow_relation_persistent_cache, follow_relation)
                pipe.execute()

    # 生成user info url
    @staticmethod
    def generate_user_info_url(token):
        return "{0}{1}{2}".format(URL_PUBLIC, token, URL_PINS)

    # 生成 follower url
    @staticmethod
    def generate_follower_info_url(token, offset, limit):
        return "{0}{1}{2}{3}{4}{5}".format(str1, token, str3, str(offset), str4, str(limit))

    # 生成 following url
    @staticmethod
    def generate_following_info_url(token, offset, limit):
        return "{0}{1}{2}{3}{4}{5}".format(str1, token, str2, str(offset), str4, str(limit))


# Token 过滤器
class TokenFilter:
    __slots__ = ('redis_connection', 'token_set_name')

    def __init__(self, redis_connection):
        # Redis 连接
        self.redis_connection = redis_connection
        # Token Set 名称
        self.token_set_name = 'tokenFilterSet'

    # 标记 Token
    def mark_token(self, token):
        self.redis_connection.sadd(self.token_set_name, token)

    # 检查 Token 是否已经存在
    def check_token(self, token):
        return self.redis_connection.sismember(self.token_set_name, token)

    # 从数据库中导入已经存在的Token
    def import_data(self):
        pass


# response 数据缓存
class ResponseBuffer:
    __slots__ = 'response_buffer'

    def __init__(self):
        self.response_buffer = queue.Queue(500)

    def get_response_from_buffer(self):
        return self.response_buffer.get()

    def put_response_to_buffer(self, response):
        self.response_buffer.put(response)
        del response



import requests
import logging
from bs4 import BeautifulSoup
from Core.Logger import log

requestHeader = {"Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,*/*;q=0.8",
                 "Accept-Encoding": "gzip, deflate, sdch, br",
                 "Accept-Language": "zh-CN,zh;q=0.8",
                 "Cache-Control": "max-age=0",
                 "Host": "www.zhihu.com",
                 "Upgrade-Insecure-Requests": "1",
                 "User-Agent": "Mozilla/5.0 (Windows NT 10.0; WOW64) AppleWebKit/537.36 (KHTML, like Gecko)"
                               " Chrome/56.0.2924.87 Safari/537.36"}

loginURL = 'https://www.zhihu.com/login/email'

mainPageURL = 'https://www.zhihu.com'

authTestURL = 'https://www.zhihu.com/api/v4/members/xzer/followers?offset=0&limit=20'


# 用户账号登陆模块
class AccountManager:
    __slots__ = ('login_token', 'password', 'auth_token', 'is_login_by_cookie', 'z_c0')

    def __init__(self, login_token, password, is_login_by_cookie, z_c0):
        # 登陆方式
        self.is_login_by_cookie = is_login_by_cookie

        # 普通登陆信息
        self.login_token = login_token
        self.password = password

        # Cookie 登陆信息
        self.z_c0 = z_c0

        # 登陆凭证（Cookies）
        self.auth_token = None

    # 返回登陆凭证
    def get_auth_token(self):
        return self.auth_token

    def login(self):
        if self.is_login_by_cookie is True:
            if log.isEnabledFor(logging.INFO):
                log.info('使用Cookie登陆方式登陆')
            return self.cookie_login()
        else:
            if log.isEnabledFor(logging.INFO):
                log.info('使用邮箱或手机号码登陆方式登陆')
            return self.common_login()

    # Cookie 登陆方式
    def cookie_login(self):
        # 创建会话
        session = requests.session()
        session.headers = requestHeader

        # 获取基本的cookie
        session.get(mainPageURL)

        # 添加用户配置的认证Cookie
        cookie = {'z_c0': self.z_c0}
        requests.utils.add_dict_to_cookiejar(session.cookies, cookie)

        # 检验是否成功登陆
        response = session.get(authTestURL)
        if response.status_code == 200:
            # 保存已经被认证Cookie
            self.auth_token = session.cookies.get_dict()
            if log.isEnabledFor(logging.INFO):
                log.info('知乎账户登陆成功')
            return True
        else:
            if log.isEnabledFor(logging.INFO):
                log.info('知乎账户登陆失败')
            return False

    # 普通登陆方式
    def common_login(self):
        # 创建会话
        session = requests.session()
        session.headers = requestHeader

        # 获取 _xsrf
        try:
            response = session.get(mainPageURL)
            input_tag = BeautifulSoup(response.text, 'html.parser').find('input', attrs={'name': '_xsrf'})
            if input_tag is None:
                return False
            _xsrf = input_tag['value']

            # login
            form_data = {'_xsrf': _xsrf,
                         'email': self.login_token,
                         'password': self.password}
            requestHeader.update({'X-Requested-With': 'XMLHttpRequest', 'X-Xsrftoken': _xsrf})
            session.headers = requestHeader
            response = session.post(url=loginURL, data=form_data)
            if response.status_code == 200:
                # 检查是否已经登陆成功
                response = session.get(authTestURL)
                if response.status_code == 200:
                    # 保存登陆认证cookie
                    self.auth_token = session.cookies.get_dict()
                    if log.isEnabledFor(logging.INFO):
                        log.info('知乎账户登陆成功')
                    return True

            # 登陆失败
            if log.isEnabledFor(logging.INFO):
                log.info('知乎账户登陆失败')
            return False
        except Exception as e:
            if log.isEnabledFor(logging.ERROR):
                log.error(e)
        finally:
            session.close()

# if __name__ == '__main__':
#     account_manager = AccountManager('xxx', 'xxx')
#     account_manager.login()
#     print(account_manager.get_auth_token())
